/* md2-int.c - Reimplementation of Ron Rivest's Message Digest Algorithm */

#include <stdio.h>

#include "alg_md2.h"
#define	OK	1
#define NOTOK	(!OK)

/* The table S given below is a permutation of 0...255 constructed from the */
/* digits of pi.  It is a ``random'' nonlinear byte substitution operatiom. */

static int      S[256] = {
    41, 46, 67, 201, 162, 216, 124, 1, 61, 54, 84, 161, 236, 240, 6, 19,
    98, 167, 5, 243, 192, 199, 115, 140, 152, 147, 43, 217, 188, 76, 130, 202,
    30, 155, 87, 60, 253, 212, 224, 22, 103, 66, 111, 24, 138, 23, 229, 18,
    190, 78, 196, 214, 218, 158, 222, 73, 160, 251, 245, 142, 187, 47, 238, 122,
    169, 104, 121, 145, 21, 178, 7, 63, 148, 194, 16, 137, 11, 34, 95, 33,
    128, 127, 93, 154, 90, 144, 50, 39, 53, 62, 204, 231, 191, 247, 151, 3,
    255, 25, 48, 179, 72, 165, 181, 209, 215, 94, 146, 42, 172, 86, 170, 198,
    79, 184, 56, 210, 150, 164, 125, 182, 118, 252, 107, 226, 156, 116, 4, 241,
    69, 157, 112, 89, 100, 113, 135, 32, 134, 91, 207, 101, 230, 45, 168, 2,
    27, 96, 37, 173, 174, 176, 185, 246, 28, 70, 97, 105, 52, 64, 126, 15,
    85, 71, 163, 35, 221, 81, 175, 58, 195, 92, 249, 206, 186, 197, 234, 38,
    44, 83, 13, 110, 133, 40, 132, 9, 211, 223, 205, 244, 65, 129, 77, 82,
    106, 220, 55, 200, 108, 193, 171, 250, 36, 225, 123, 8, 12, 189, 177, 74,
    120, 136, 149, 139, 227, 99, 232, 109, 233, 203, 213, 254, 59, 0, 29, 57,
    242, 239, 183, 14, 102, 88, 208, 228, 166, 119, 114, 248, 235, 117, 75, 10,
    49, 68, 80, 180, 143, 237, 31, 26, 219, 153, 141, 51, 159, 17, 131, 20,
};

/* ARGSUSED */
int
md2_init_hash(hash, parm, key)
	struct MD2Hash **hash;
	struct MD2Parameters *parm;
	struct MD2Key  *key;
{
	struct MD2Hash *result;

	result = (struct MD2Hash *) calloc(1, sizeof(*result));
	if (result == NULLMD2HASH)
		return (NOTOK);

	bzero(result->D, 16);
	result->L = 0;

	(*hash) = result;

	return (OK);
}

/* ARGSUSED */
int
md2_free_hash(hash, parm)
	struct MD2Hash *hash;
	struct MD2Parameters *parm;
{
	free((char *) hash);

	return (OK);
}

/* ARGSUSED */
int
md2_compare_hash(h1, h2, key, parm)
	struct MD2Hash *h1;
	struct MD2Hash *h2;
	struct MD2Key *key;
	struct MD2Parameters *parm;
{
	int             i;
	int             result;

	result = OK;
	for (i = 0; i < 16; i++)
		if (h1->D[i] != h2->D[i])
			result = NOTOK;

	return (result);
}

/* ARGSUSED */
int
md2_get_blocksize(parm)
       struct MD2Parameters  *parm;
{
	return (128);
}

/* ARGSUSED */
int
md2_get_hashsize(parm)
	struct MD2Parameters  *parm;
{
	return (128);
}

/* ARGSUSED */
int
md2_hash_block(hash, cp, key, parm)
	struct MD2Hash *hash;
	char           *cp;
	struct MD2Key  *key;
	struct MD2Parameters *parm;
{
	char            D[48];
	int             l;
	int             i;
	int             j;
	int             t;
	char           *ptr;

	l = hash->L;

	for (i = 0; i < 16; i++) {
		D[i] = (hash->D)[i];
		D[i + 16] = cp[i];
		D[i + 32] = (hash->D)[i] ^ cp[i];
		l = ((hash->C)[i] ^= S[0xFF & (cp[i] ^ l)]);
	}

	t = 0;
	for (j = 0; j < 18; j++) {
		ptr = D;
		for (i = 0; i < 48; i++, ptr++)
			t = (*ptr ^= S[t]) & 255;
		t = (t + j) & 255;
	}

	bcopy(D, hash->D, 16);
	hash->L = l;

	return (OK);
}

int
md2_hash_short_block(hash, buff, extra, key, parm)
	struct MD2Hash *hash;
	char           *buff;
	int             extra;
	struct MD2Key  *key;
	struct MD2Parameters *parm;
{
	int             i;
	char            c[16];

	for (i = 0; i < extra; i++)
		c[i] = buff[i];
	for (i = extra; i < 16; i++)
		c[i] = (16 - extra);
	return md2_hash_block(hash, c, key, parm);
}

int
md2_terminate_hash(hash, key, parms)
	struct MD2Hash *hash;
	struct MD2Key  *key;
	struct MD2Parameters *parms;
{
	char            buff[16];

	bcopy(hash->C, buff, 16);
	(void) md2_hash_block(hash, buff, key, parms);

	return (OK);
}

/* ARGSUSED */
int
md2_pack_hash(in, out, n_bits, parms)
	struct MD2Hash *in;
	char          **out;
	int            *n_bits;
	struct MD2Parameters *parms;
{
	*out = (char *) calloc(16, 1);
	if (*out == (char *) 0)
		return (NOTOK);
	*n_bits = 128;
	bcopy(in->D, *out, 16);
	return (OK);
}

/* ARGSUSED */
int
md2_unpack_hash(in, out, parms)
	char           *in;
	struct MD2Hash **out;
	struct MD2Parameters *parms;
{
	*out = (struct MD2Hash *) calloc(1, sizeof(struct MD2Hash));
	if (*out == (struct MD2Hash *) 0)
		return (NOTOK);
	bcopy(in, (*out)->D, 16);
	return (OK);
}

int
md2_display(hash)
	struct MD2Hash *hash;
{
	int             i;

	(void) printf("D = ");
	for (i = 0; i < 16; i++)
		(void) printf("%02x", hash->D[i] & 255);
	(void) printf("\n");
}
