/* md4a-int.c - The original MD4 Message Digest Algorithm.
 * This is now obsolete, as the MD4 specification has changed.
 * Is is preserved as a separate file in case we want to run the
 * original algorithm.
 */

#include <stdio.h>

#include "general.h"
#include "osisec/alg_md4a.h"

/* ARGSUSED */
int
md4a_init_hash(hash, parms, key)
	struct MD4AHash **hash;
	struct MD4AParameters *parms;
	struct MD4AKey *key;
{
	struct MD4AHash *result;

	result = (struct MD4AHash *) calloc(1, sizeof(*result));

	result->A = 0xc3658a99;
	result->B = 0xc0f6281e;
	result->C = 0xa9605167;
	result->D = 0x24224167;

	*hash = result;

	return (OK);
}

#define F(x, y, z) (((x) & (y)) | ((~(x)) & (z)))
#define G(x, y, z) (((x) & (y)) | ((x) & (z)) | ((y) & (z)))
#define H(x, y, z) ((x)^(y)^(z))

#define ROT(x, s) (((x) << (s)) | ((x) >> (32 - (s))))

#define R1(a, b, c, d, i, s) hash->a = \
ROT((hash->a + F(hash->b, hash->c, hash->d) + text[i]), s)

#define R2(a, b, c, d, i, s) hash->a = \
ROT((hash->a + G(hash->b, hash->c, hash->d) + text[i] + 0xa352b7f9), s)

#define R3(a, b, c, d, i, s) hash->a = \
ROT((hash->a + H(hash->b, hash->c, hash->d) + text[i] + 0x6249bf17), s)

/* ARGSUSED */
int
md4a_hash_block(hash, text, key, parms)
	struct MD4AHash *hash;
	char           *text;
	struct MD4AKey *key;
	struct MD4AParameters *parms;
{
	R1(A, B, C, D, 0, 3);
	R1(D, A, B, C, 1, 7);
	R1(C, D, A, B, 2, 11);
	R1(B, C, D, A, 3, 19);
	R1(A, B, C, D, 4, 3);
	R1(D, A, B, C, 5, 7);
	R1(C, D, A, B, 6, 11);
	R1(B, C, D, A, 7, 19);
	R1(A, B, C, D, 8, 3);
	R1(D, A, B, C, 9, 7);
	R1(C, D, A, B, 10, 11);
	R1(B, C, D, A, 11, 19);
	R1(A, B, C, D, 12, 3);
	R1(D, A, B, C, 13, 7);
	R1(C, D, A, B, 14, 11);
	R1(B, C, D, A, 15, 19);

	R2(A, B, C, D, 0, 3);
	R2(D, A, B, C, 4, 5);
	R2(C, D, A, B, 8, 9);
	R2(B, C, D, A, 12, 13);
	R2(A, B, C, D, 1, 3);
	R2(D, A, B, C, 5, 5);
	R2(C, D, A, B, 9, 9);
	R2(B, C, D, A, 13, 13);
	R2(A, B, C, D, 2, 3);
	R2(D, A, B, C, 6, 5);
	R2(C, D, A, B, 10, 9);
	R2(B, C, D, A, 14, 13);
	R2(A, B, C, D, 3, 3);
	R2(D, A, B, C, 7, 5);
	R2(C, D, A, B, 11, 9);
	R2(B, C, D, A, 15, 13);

	R3(A, B, C, D, 0, 3);
	R3(D, A, B, C, 8, 9);
	R3(C, D, A, B, 4, 11);
	R3(B, C, D, A, 12, 15);
	R3(A, B, C, D, 2, 3);
	R3(D, A, B, C, 10, 9);
	R3(C, D, A, B, 6, 11);
	R3(B, C, D, A, 14, 15);
	R3(A, B, C, D, 1, 3);
	R3(D, A, B, C, 9, 9);
	R3(C, D, A, B, 5, 11);
	R3(B, C, D, A, 13, 15);
	R3(A, B, C, D, 3, 3);
	R3(D, A, B, C, 11, 9);
	R3(C, D, A, B, 7, 11);
	R3(B, C, D, A, 15, 15);

	return (OK);
}

int
md4a_display(hash)
	struct MD4AHash *hash;
{
	(void) printf("A = %04x\n", hash->A);
	(void) printf("B = %04x\n", hash->B);
	(void) printf("C = %04x\n", hash->C);
	(void) printf("D = %04x\n", hash->D);

	return (OK);
}


/* ARGSUSED */
int
md4a_compare_hash(h1, h2, key, parms)
	struct MD4AHash *h1, *h2;
	struct MD4AKey *key;
	struct MD4AParameters *parms;
{

	if (h1->A != h2->A)
		return (NOTOK);
	if (h1->B != h2->B)
		return (NOTOK);
	if (h1->C != h2->C)
		return (NOTOK);
	if (h1->D != h2->D)
		return (NOTOK);

	return (OK);
}

/* ARGSUSED */
int
md4a_get_blocksize(parms)
	struct MD4AParameters *parms;
{
	return (128);
}

/* ARGSUSED */
int
md4a_get_hashsize(parm)
	struct MD4AParameters *parm;
{
	return (128);
}


/* ARGSUSED */
int
md4a_terminate_hash(hash, key, parms)
	struct MD4AHash *hash;
	struct MD4AKey *key;
	struct MD4AParameters *parms;
{
	return (OK);
}

int
md4a_hash_short_block(hash, buff, extra, key, parm)
	struct MD4AHash *hash;
	char           *buff;
	int             extra;
	struct MD4AKey *key;
	struct MD4AParameters *parm;
{
	int             i;
	int             rc;

	/* Note the right padding convention */

	for (i = extra; i < 16; i++)
		buff[i] = (char) 0;

	rc = md4a_hash_block(hash, buff, key, parm);

	return (rc);
}


/* ARGSUSED */
int
md4a_pack_hash(hash, out, n_bits, parm)
	struct MD4AHash *hash;
	char          **out;
	int            *n_bits;
	struct MD4AParameters *parm;
{
	char           *cp;

	cp = calloc(1, 16);
	if (cp == NULLCP)
		return (NOTOK);

	cp[0] = hash->D & 255;
	cp[1] = (hash->D >> 8) & 255;
	cp[2] = (hash->D >> 16) & 255;
	cp[3] = (hash->D >> 24) & 255;

	cp[4] = hash->C & 255;
	cp[5] = (hash->C >> 8) & 255;
	cp[6] = (hash->C >> 16) & 255;
	cp[7] = (hash->C >> 24) & 255;

	cp[8] = hash->B & 255;
	cp[9] = (hash->B >> 8) & 255;
	cp[10] = (hash->B >> 16) & 255;
	cp[11] = (hash->B >> 24) & 255;

	cp[12] = hash->A & 255;
	cp[13] = (hash->A >> 8) & 255;
	cp[14] = (hash->A >> 16) & 255;
	cp[15] = (hash->A >> 24) & 255;

	(*n_bits) = 128;
	(*out) = cp;

	return (OK);
}


int
md4a_unpack_hash(cp, hash)
	unsigned char  *cp;
	struct MD4AHash **hash;
{
	struct MD4AHash *result;

	result = (struct MD4AHash *) calloc(1, sizeof(*result));
	if (result == (struct MD4AHash *) 0)
		return (NOTOK);

	result->D = cp[0] | (cp[1] << 8) | (cp[2] << 16) | (cp[3] << 24);
	result->C = cp[4] | (cp[5] << 8) | (cp[6] << 16) | (cp[7] << 24);
	result->B = cp[8] | (cp[9] << 8) | (cp[10] << 16) | (cp[11] << 24);
	result->A = cp[12] | (cp[13] << 8) | (cp[14] << 16) | (cp[15] << 24);

	(*hash) = result;

	return (OK);
}

/* ARGSUSED */
int
md4a_free_hash(hash, parm)
	struct MD4AHash *hash;
	struct MD4AParameters *parm;
{
	free((char *) hash);
	return (OK);
}

