/* gcd.c - Fast Binary Greatest Commmon Divisor */

#include <stdio.h>
#include "numalgs.h"

static int 
zero(a, l)
	unsigned        a[];
	int             l;
{
	int             i;
	for (i = LSW; i <= l; i++)
		if (a[i] != 0)
			return (0);
	return (-1);
}

static int 
odd(a)
	unsigned        a[];
{
	if ((a[LSW] & 1) == 0)
		return 0;
	else
		return -1;
}

static int 
even(a)
	unsigned        a[];
{
	if ((a[LSW] & 1) == 0)
		return -1;
	else
		return 0;
}

/*  */

/**********************************************************
 * At all times in fastgcd(),                             *
 *  u.u1 - v.u2 = u3                                      *
 * -u.v1 + v.v2 = v3                                      *
 * These equations are added and subtracted according to  *
 * the Euclidean algorithm, and also divided by 2 where   *
 * possible.                                              *
 **********************************************************/

fastgcd(u, u1, u2, u3, v, v1, v2, v3, l)
	unsigned        u[];
	unsigned        u1[];	/* 1  */
	unsigned        u2[];	/* 0  */
	unsigned        u3[];	/* u  */
	unsigned        v[];
	unsigned        v1[];	/* 0  */
	unsigned        v2[];	/* 1  */
	unsigned        v3[];	/* v  */
	int             l;
{
	unsigned        cu1;
	unsigned        cu2;
	unsigned        cv1;
	unsigned        cv2;
	void		exit();

	clear(u1, l);
	u1[LSW] = 1;		/* u1 <- 1 */
	clear(u2, l);		/* u2 <- 0 */
	clear(v1, l);		/* v1 <- 0 */
	clear(v2, l);		/* v2 <- 1 */
	v2[LSW] = 1;
	cu1 = 0;
	cu2 = 0;
	cv1 = 0;
	cv2 = 0;

	if (zero(u, l)) {
		(void)fprintf(stderr, "Zero gcd.\n");
		exit(1);
	}
	if (zero(v, l)) {
		(void)fprintf(stderr, "Zero gcd.\n");
		exit(1);
	}
	while (even(u) && even(v)) {	/* Wlog. u or v is odd. */
		bsr(u, u, l, 1);
		bsr(v, v, l, 1);
	}

	copy(u3, u, l);
	copy(v3, v, l);

#ifdef DEBUG
	(void)fprintf(stderr, "\n");
	(void)fprintf(stderr, "u1 = (%d) ", cu1);
	fhexdump(stderr, u1, l);
	(void)fprintf(stderr, "u2 = (%d) ", cu2);
	fhexdump(stderr, u2, l);
	(void)fprintf(stderr, "u3 = ");
	fhexdump(stderr, u3, l);
	(void)fprintf(stderr, "\n");
	(void)fprintf(stderr, "v1 = (%d) ", cv1);
	fhexdump(stderr, v1, l);
	(void)fprintf(stderr, "v2 = (%d) ", cv2);
	fhexdump(stderr, v2, l);
	(void)fprintf(stderr, "v3 = ");
	fhexdump(stderr, v3, l);
#endif

	while (!zero(u3, l) && !zero(v3, l)) {
		if (even(u3)) {
			while (even(u3)) {
				if (odd(u1) || odd(u2)) {	/* Can't both be odd. */
					cu1 += add(u1, v, l);	/* Make both even.    */
					cu2 += add(u2, u, l);
				}
#ifdef DEBUG
				(void)fprintf(stderr, "\n");
				(void)fprintf(stderr, "u1 = (%d) ", cu1);
				fhexdump(stderr, u1, l);
				(void)fprintf(stderr, "u2 = (%d) ", cu2);
				fhexdump(stderr, u2, l);
				(void)fprintf(stderr, "u3 = ");
				fhexdump(stderr, u3, l);
#endif
				bsr(u1, u1, l, 1);	/* Halve all three.   */
				if (cu1 & 1)
					u1[l] |= TOPB;
				cu1 = cu1 >> 1;
				bsr(u2, u2, l, 1);
				if (cu2 & 1)
					u2[l] |= TOPB;
				cu2 = cu2 >> 1;
				bsr(u3, u3, l, 1);
			}
		} else {
			while (even(v3)) {
				if (odd(v1) || odd(v2)) {
					cv1 += add(v1, v, l);
					cv2 += add(v2, u, l);
				}
#ifdef DEBUG
				(void)fprintf(stderr, "\n");
				(void)fprintf(stderr, "v1 = (%d) ", cv1);
				fhexdump(stderr, v1, l);
				(void)fprintf(stderr, "v2 = (%d) ", cv2);
				fhexdump(stderr, v2, l);
				(void)fprintf(stderr, "v3 = ");
				fhexdump(stderr, v3, l);
#endif
				bsr(v1, v1, l, 1);
				if (cv1 & 1)
					v1[l] |= TOPB;
				cv1 = cv1 >> 1;
				bsr(v2, v2, l, 1);
				if (cv2 & 1)
					v2[l] |= TOPB;
				cv2 = cv2 >> 1;
				bsr(v3, v3, l, 1);
			}
		}
#ifdef DEBUG
		(void)fprintf(stderr, "Next round...\n");
#endif
		if (cmp(u3, v3, l) > 0) {
			cu1 += add(u1, v1, l) + cv1;
			cu2 += add(u2, v2, l) + cv2;
			(void) sub(u3, v3, l);
		} else {
			cv1 += add(v1, u1, l) + cu1;
			cv2 += add(v2, u2, l) + cu2;
			(void) sub(v3, u3, l);
		}
	}
	/* when u1 > v , modified by kikn */
	if((cu1 > 0) || (cmp(u1, v, l) > 0)){
		while((cu1 += sub(u1, v, l)) > 0){
			;
		}
	}
}
