/* kgen.c - Generate strong RSA key. */

#include <stdio.h>

#include "numalgs.h"
#include "pem.h"

/* Must initialise with srandom()  */
large_random(a, l)
	unsigned        a[];
	int             l;
{
	int             i;
	static int      f=0;
	long		random(), r;

/*	for (i = LSW; i <= l; i++)
		a[i] = ( random() ) & MASK;*/
/* modified by kuroda */
        for (i = LSW; i <= l; i++){
                if (f == 0)     { r = random(); f = 1; }
                else            { r = r >> 16;  f = 0; }
                a[i] = r & MASK;
        }

}

unsigned       *
ualloc(n)
	int             n;
{
	unsigned       *result;

	result = (unsigned *)
		calloc((unsigned) n + 2, sizeof(unsigned)); /* should be +1 */

	return result;
}

int 
trialdiv(a, l, n)
	unsigned        a[];
	int             l;
	register unsigned n;
{
	register int    i;
	register unsigned cy;

	cy = 0;
	for (i = l; i >= 0; i--)
		cy = ((cy << BITS) + a[i]) % n;
	return (cy);
}

int 
primesieve(a, l, sieve)
	unsigned        a[];
	int             l;
	unsigned        sieve[];
{
	int             i;

	for (i = 1; i <= sieve[0]; i++)
		if (trialdiv(a, l, sieve[i]) == 0)
			return (0);
	return (-1);
}

void 
prime(a, l, s, sieve)
	unsigned        a[];
	int             l;
	int             s;
	unsigned        sieve[];
{
	int             found;
	unsigned       *q;
	unsigned       *x;
	unsigned       *y;
	unsigned       *u;
	unsigned       *v;

	found = 0;
	q = ualloc(l);
	x = ualloc(l);
	y = ualloc(l);
	u = ualloc(l);
	v = ualloc(l);

	while (found == 0) {
		large_random(a, l);
		a[LSW] = a[LSW] | 1;	/* Force to be even */
		bsl(a, a, l, s);/* Shift  */
		a[l] = a[l] | TOPB;	/* Force most sig. bit high */

#ifdef DEBUG
		(void)fprintf(stderr, "Trying ");
		fhexdump(stderr, a, l);
		fflush(stderr);
#endif

		found = primesieve(a, l, sieve);
		if (found)
			found = pptest(a, l, s, q, u, v, x, y);
	}
}

/*****************************************************************
 *                                                               *
 *  Find a 'strong' prime, ie. p-1 has a large prime factor.     *
 *                                                               *
 *****************************************************************/

void 
strongprime(r, lr, sr, t, lt, st, sieve)
	unsigned        r[];
	int             lr;
	int             sr;
	unsigned        t[];
	int             lt;
	int             st;
	unsigned        sieve[];
{
	int             found;
	int             disp;
	int             i;
	unsigned       *q;
	unsigned       *u;
	unsigned       *v;
	unsigned       *x;
	unsigned       *y;

	q = ualloc(lr);
	u = ualloc(lr);
	v = ualloc(lr);
	x = ualloc(lr);
	y = ualloc(lr);

	clear(t, lr);		/* The lt+1 to lr entries will stay
				 * cleared... */
	prime(t, lt, st, sieve);
	disp = lr - lt;
	clear(r, lr);
	for (i = LSW; i <= lt; i++)
		r[i + disp] = t[i];

#ifdef DEBUG
	if (!(r[lr] & TOPB))
		(void)fprintf(stderr, "Alignment failed.\n");
#endif

	(void)ads(r, lr, (unsigned)(1 << sr));	/* r <- kt + 1  */
	bsr(t, t, lt, st);	/* Align t with r */
	bsl(t, t, lr, sr);
	bsl(t, t, lr, 1);	/* t <- 2t */

	found = 0;
	while (found == 0) {
#ifdef DEBUG
		(void)fprintf(stderr, "S.Trying ");
		fhexdump(stderr, r, lr);
		fflush(stderr);
#endif

		found = primesieve(r, lr, sieve);
		if (found != 0)
			found = pptest(r, lr, sr, q, u, v, x, y);
		if (found == 0)
			(void) add(r, t, lr);
	}
}

/*****************************************************************
 *                                                               *
 *                Probabilistic Primality Test.                  *
 *                                                               *
 *   Each time round the loop, non-primes are rejected with a    *
 * probability of at least 1/4, independant of the previous time *
 * round. The first iteration lets through non-primes with a     *
 * probability of about 1 in 10,000.                             *
 *   Note that this test depends upon the Riemann Hypothesis.    *
 *****************************************************************/

int 
pptest(a, l, s, q, u, v, x, y)
	unsigned        a[];
	int             l;
	int             s;
	unsigned        q[];
	unsigned        u[];
	unsigned        v[];
	unsigned        x[];
	unsigned        y[];
{
	int             k;
	int             j;
	int             tries;
	int             found;

	bsr(q, a, l, s + 1);	/* Calculate Knuth's parameters */
	k = 1;
	while ((q[LSW] & 1) == 0) {
		bsr(q, q, l, 1);
		k++;
	}

	tries = 0;
	found = -1;
	while (found && (tries < 4)) {
		tries++;
		large_random(x, l);	/* Random x for Prob. Primality Test */
		reduce(x, 0, a, l);	/* n > x > 2 */
		modexp(y, x, q, a, l, s, u, v);	/* r <- x ^ q (mod a) */
		j = 0;
		while ((j < k) & (cps(y, l, 1 << s) != 0) & (cpdelta(y, a, 1 << s, l) != 0)) {
			bsr(u, y, l, s);
			copy(v, y, l);
			modmul(y, u, v, a, l);	/* y <- y ^ 2 (mod a) */
			j++;
		}

		/**********************************************************
	         * if x^q = 1, then x^(a-1) = 1, so probably prime.       *
	         * if x^(q.2^j) = -1, x^(a-1) =1, so probably prime.      *
	         * if x^(q.2^j) = 1, we know x^(q.2^j-1) <> -1            *
	         *  hence square roots are not bi-unique, so a not prime. *
	         **********************************************************/

		if ((j != 0) && (cps(y, l, 1 << s) != 0))
			found = 0;
	}

#ifdef DEBUG
	(void)fprintf(stderr, "tries = %d\n", tries);
	fflush(stderr);
#endif

	return (found);
}

/********************************************************
 *                                                      *
 *     Generate parameters needed by RSA routines.      *
 *                                                      *
 ********************************************************/

void 
rsaparms(p, q, n, d, e, a, x, y, l, ll)
	unsigned        p[];	/* First factor of modulus */
	unsigned        q[];	/* Second factor           */
	unsigned        n[];	/* The public modulus      */
	unsigned        d[];	/* Public exponent         */
	unsigned        e[];	/* Secret exponent         */
	unsigned        a[];	/* p^-1 (mod q)            */
	unsigned        x[];	/* e (mod p)               */
	unsigned        y[];	/* e (mod q)               */
	int             l;	/* MSW of a factor         */
	int             ll;	/* MSW of n                */
{
	unsigned       *w1, *w2, *w3, *v1, *v2, *v3, *phi;
	int             shift;
	void 		exit();

	w1 = ualloc(ll);
	w2 = ualloc(ll);
	w3 = ualloc(ll);
	v1 = ualloc(ll);
	v2 = ualloc(ll);
	v3 = ualloc(ll);
	phi = ualloc(ll);

	multiply(n, p, q, l);	/* n <- p.q */

	copy(w1, p, l);
	(void)sbs(w1, l, 1);
	copy(w2, q, l);
	(void)sbs(w2, l, 1);
	multiply(phi, w1, w2, l);	/* phi is Euler's function, phi(n) */

#ifdef DEBUG
	(void)fprintf(stderr, "phi = ");
	fhexdump(stderr, phi, ll);
#endif

	copy(w1, d, ll);
	fastgcd(w1, e, w2, w3, phi, v1, v2, v3, ll + 1);	/* w1,w2,v1,v2,v3 used
								 * as scratch */
#ifdef DEBUG
	(void)fprintf(stderr, "e = ");
	fhexdump(stderr, e, ll);
	(void)fprintf(stderr, "d = ");
	fhexdump(stderr, d, ll);
#endif

	if (cps(w3, l, 1) != 0) {
		(void)fprintf(stderr, "RSAPARMS Failure: E and phi not coprime.\n");
		exit(1);
	}
#ifdef DEBUG
	(void)fprintf(stderr, "phi = ");
	fhexdump(stderr, phi, ll);
#endif

	copy(w1, d, ll);
	copy(w2, e, ll);
	modmul(w3, w1, w2, phi, ll);

#ifdef DEBUG
	(void)fprintf(stderr, "d.e mod phi = ");
	fhexdump(stderr, w3, ll);
#endif

	shift = bitsfree(p, l);
	bsl(w1, p, l, shift);
	(void)sbs(w1, l, (unsigned)(1 << shift));
	if (bitsfree(e, ll) < shift)
		(void)fprintf(stderr, "e too large.\n");
	bsl(w2, e, ll, shift);	/* DANGER - carry? */
	fullreduce(w2, ll - l, w1, l);	/* x <- e (mod p-1) */
	bsr(x, w2, l, shift);

	shift = bitsfree(q, l);
	bsl(w1, q, l, shift);
	(void)sbs(w1, l, (unsigned)(1 << shift));
	if (bitsfree(e, ll) < shift)
		(void)fprintf(stderr, "e too large.\n");
	bsl(w2, e, ll, shift);
	fullreduce(w2, ll - l, w1, l);	/* y <- e (mod q-1) */
	bsr(y, w2, l, shift);

	fastgcd(p, a, w2, w3, q, v1, v2, v3, l);	/* a <- p^-1 (mod q) */

#ifdef DEBUG
	(void)fprintf(stderr, "u3 = ");
	fhexdump(stderr, w3, l);
	(void)fprintf(stderr, "v3 = ");
	fhexdump(stderr, v3, l);
#endif
}
