/* modular.c - Routines to perform modular arithmetic */

#include <stdio.h>
#include "numalgs.h"

/*  */

void 
copy(a, b, l)
	unsigned        a[];
	unsigned        b[];
	int             l;
{
	register int    i;

	for (i = 0; i <= l; i++)
		a[i] = b[i];
}

void 
clear(a, l)
	unsigned        a[];
	int             l;
{
	register int    i;
	register unsigned *ap;

	ap = a;
	i = l + 1;

	while (i--) {
		(*ap) = 0;
		ap++;
	}
}

int 
cmp(a, b, l)
	unsigned        a[];
	unsigned        b[];
	int             l;
{
	int             i;

	for (i = l; i >= LSW; i--) {
		if (a[i] > b[i])
			return (1);
		if (a[i] < b[i])
			return (-1);
	}
	return (0);
}

int 
cpdelta(a, b, l, cy)
	unsigned        a[];
	unsigned        b[];
	int             l;
	int             cy;
{
	int             i;

	for (i = LSW; i <= l; i++) {
		cy = cy + a[i] - b[i];
		if ((cy & MASK) != 0)
			return (-1);
		cy = cy >> BITS;
	}
	return (0);
}

int 
cps(a, l, n)
	unsigned        a[];
	int             l;
	int             n;
{
	int             i;
	int             cy;

	cy = 0;
	for (i = l; i > LSW; i--)
		if (a[i] != 0) {
			cy = -1;
			break;
		}
	if (cy == 0)
		cy = (a[LSW] - n);
	return (cy);
}

int 
add(a, b, l)
	unsigned        a[];
	unsigned        b[];
	int             l;
{
	register int    i;
	register unsigned cy;

	cy = 0;
	for (i = LSW; i <= l; i++) {
		cy = cy + a[i] + b[i];
		a[i] = cy & MASK;
		cy = cy >> BITS;
	}
	return cy;
}

unsigned 
adn(a, b, l, n)
	unsigned        a[];
	unsigned        b[];
	int             l;
	register unsigned n;
{
	register int    i;
	register unsigned cy;
	register unsigned *ap;
	register unsigned *bp;

	cy = 0;
	ap = a;
	bp = b;
	i = l + 1;

	while (i--) {
		cy = cy + (*ap) + (*bp) * (n);
		(*ap) = cy & MASK;
		ap++;
		bp++;
		cy = cy >> BITS;
	}
	return (cy);
}

int 
ads(a, l, n)
	unsigned        a[];
	int             l;
	unsigned        n;
{
	register int    i;
	register unsigned cy;

	cy = n;
	for (i = LSW; i <= l; i++) {
		if (cy == 0)
			break;
		cy = cy + a[i];
		a[i] = cy & MASK;
		cy = cy >> BITS;
	}
	return cy;
}

int 
shl(a, l)
	unsigned        a[];
	int             l;
{
	register int    i;
	register int    cy;

	cy = a[l];
	for (i = l; i > 0; i--)
		a[i] = a[i - 1];
	a[0] = 0;
	return cy;
}

int 
sub(a, b, l)
	unsigned        a[];
	unsigned        b[];
	int             l;
{
	int             i;
	int             cy;

	cy = 0;
	for (i = 0; i <= l; i++) {
		cy = cy + a[i] - b[i];
		a[i] = cy & MASK;
		cy = cy >> BITS;
	}
	return (cy);
}

int 
sbn(a, carry, b, l, n)
	unsigned        a[];
	int             carry;
	unsigned        b[];
	int             l;
	register int    n;
{
	register int    i;
	register int    cy;
	register unsigned *ap;
	register unsigned *bp;

	cy = 0;
	ap = a;
	bp = b;
	i = l + 1;
	while (i--) {
		cy = cy + (int) (*ap) - (int) (*bp) * n;
		(*ap) = cy & MASK;
		ap++;
		bp++;
		cy = cy >> BITS;
		if (cy > 0)
			cy = cy - (MASK + 1);
	}
	cy = cy + carry;
	return (cy);
}

int 
sbs(a, l, n)
	unsigned        a[];
	int             l;
	unsigned        n;
{
	register int    i;
	register int    cy;

	cy = -n;
	for (i = LSW; i <= l; i++) {
		if (cy == 0)
			break;
		cy = cy + a[i];
		a[i] = cy & MASK;
		cy = cy >> BITS;
	}
	return cy;
}

int 
sbl(a, carry, b, l)
	unsigned        a[];
	int             carry;
	unsigned        b[];
	int             l;
{
	register int    cy;
	register int    i;

	cy = 0;
	for (i = LSW; i < l; i++) {
		cy = cy + a[i + 1] - b[i];
		a[i] = cy & MASK;
		cy = cy >> BITS;
	}
	cy = cy + carry - b[l];
	return (cy);
}

void 
multiply(a, b, c, l)
	unsigned        a[];
	unsigned        b[];
	unsigned        c[];
	int             l;
{
	int             i;
	unsigned        cy;

	clear(a, 2 * l + 1);
	for (i = LSW; i <= l; i++) {
		cy = adn(a + i, b, l, c[i]);
		(void) ads(a + l + i + 1, l - i, cy);
	}
}

void 
readf(f, a, l)
	FILE           *f;
	unsigned        a[];
	int             l;
{
	int             i;
	int             j;
	int             n;

	for (i = l; i >= LSW; i--) {
		a[i] = 0;
		for (j = 0; j < 4; j++) {
			(void)fscanf(f, "%x:1", &n);
			a[i] = (a[i] << 4) | n;
		}
	}
}

void 
hexdump(a, l)
	unsigned        a[];
	int             l;
{
	int             i, j;
	unsigned        n;

	for (i = l; i >= LSW; i--) {
		n = a[i];
		for (j = 0; j < BITS; j = j + 4) {
			(void)printf("%x", n >> HNIB);
			n = (n << 4) & MASK;
		}
	}
	(void)printf("\n");
}

void 
fhexdump(f1, a, l)
	FILE	       *f1;
	unsigned       *a;
	int             l;
{
	int             i, j;
	unsigned        n;

	for (i = l; i >= LSW; i--) {
		n = a[i];
		for (j = 0; j < BITS; j = j + 4) {
			(void)fprintf(f1, "%x", n >> HNIB);
			n = (n << 4) & MASK;
		}
	}
	(void)fprintf(f1, "\n");
}

/*****************************************************************
 *                 Modular Reduction Routines                    *
 *                                                               *
 * These require that the modulus has the MSB of the MSW high    *
 * If necessary, all numbers are shifted left by an appropriate  *
 * number of bits in order to acheive this.                      *
 *****************************************************************/

void 
reduce(a, carry, b, l)
	unsigned        a[];
	unsigned        carry;
	unsigned        b[];
	int             l;
{
	register unsigned qhat;
	register unsigned temp;
	int             cy;

	temp = (carry << BITS) + a[l];
	qhat = temp / b[l];

	temp = temp - qhat * b[l];	/* This is >=0, < b[l] */

	temp = (temp << BITS) + a[l - 1];

	if (temp < qhat * b[l - 1])
		qhat = qhat - 1;

	if (qhat == MASK + 1)
		qhat = qhat - 1;/* Do this case in 2 steps */

#ifdef DEBUG
	if (qhat > MASK) {
		(void)fprintf(stderr, "qhat = %d\n", qhat);	/* This is almost certainly
						 * an error */
		(void)fprintf(stderr, "Carry=%d\n", carry);
		(void)fprintf(stderr, "a[l]=%d\n", a[l]);
		carry = sbl(a, carry, b, l);
		qhat = qhat & MASK;
	}
#endif

	if (qhat != 0)
/*		cy = sbn(a, carry, b, l, qhat);		 PW chg for SABER */
		cy = sbn(a, (int)carry, b, l, (int)qhat);
	else
		cy = 0;

	if (cy == -1)
		(void) add(a, b, l);

#ifdef DEBUG
	if (cy == 1)
		(void) sbn(a, cy, b, l, 1);
	if ((cy != 0) && (cy != -1))
		(void)fprintf(stderr, "Unusual carry %d\n", cy);
#endif
}

void 
fullreduce(a, k, b, l)
	unsigned        a[];
	int             k;
	unsigned        b[];
	int             l;
{
	int             i;
	unsigned        cy;

	cy = 0;
	for (i = k; i >= 0; i--) {
		reduce(a + i, cy, b, l);
		cy = a[i + l];
	}
}


void 
bsr(a, b, l, s)
	unsigned        a[];
	unsigned        b[];
	int             l;
	int             s;
{
	register int    i;

	for (i = 0; i < l; i++)
		a[i] = (((b[i + 1] << BITS) | b[i]) >> s) & MASK;
	a[l] = (b[l] >> s) & MASK;
}

void 
bsl(a, b, l, s)
	unsigned        a[];
	unsigned        b[];
	int             l;
	int             s;
{
	register int    i;

	for (i = l; i >= 1; i--)
		a[i] = (((b[i] << BITS) | b[i - 1]) >> (BITS - s)) & MASK;
	a[0] = (b[0] << s) & MASK;
}

int 
bitsfree(a, l)
	unsigned        a[];
	int             l;
{
	int             i, x;

	x = a[l];
	if (x == 0)
		return (BITS);
	i = 0;
	while ((x & TOPB) == 0) {
		x = x << 1;
		i++;
	}
	return (i);
}

void 
modmul(result, a, b, modulus, l)
	unsigned        result[];
	unsigned        a[];
	unsigned        b[];
	unsigned        modulus[];
	int             l;
{
	register int    i;
	register unsigned cy;

	clear(result, l);
	for (i = LSW; i <= l; i++) {
		cy = adn(result, a, l, b[i]);
		reduce(result, cy, modulus, l);
		cy = shl(a, l);
		reduce(a, cy, modulus, l);
	}
}

void 
tablemul(result, a, table, modulus, l)
	unsigned        result[];
	unsigned        a[];
	unsigned      **table;
	unsigned       *modulus;
	int             l;
{
	int             i;
	unsigned        cy;

	clear(result, l);
	for (i = LSW; i <= l; i++) {
		cy = adn(result, table[i], l, a[i]);
		reduce(result, cy, modulus, l);
	}
}


/********************************************************************
 *                    Modular Exponentiation                        *
 *                                                                  *
 * x ^ (a + 2b + 4c + ...) = (x^a)(x^2^b)(x^2^2^c) ...              *
 *                         = (((x^c)^2) (x^b))^2) (x^a) ...         *
 *                                                                  *
 * The former is the traditional n^3 time exponentiation algorithm. *
 * Note that in the latter expansion multiplication is always by x. *
 * This enables the calculation to be speeded up by precalculating  *
 * multiples of x. - This is used in fastexp().                     *
 ********************************************************************/

void 
modexp(r, x, p, m, l, shift, w1, w2)
	unsigned        r[];
	unsigned        x[];
	unsigned        p[];
	unsigned        m[];
	int             l;
	int             shift;
	unsigned        w1[];
	unsigned        w2[];
{
	int             i;
	int             j;
	int             b;

	for (i = LSW; i <= l; i++)
		r[i] = 0;
	r[LSW] = 1 << shift;	/* r <- 1     */

	for (i = l; i >= LSW; i--) {
		b = p[i];
		for (j = 0; j < BITS; j++) {
			copy(w1, r, l);
			bsr(w2, r, l, shift);
			modmul(r, w1, w2, m, l);	/* r <- r . r */
			if (b & TOPB) {
				copy(w1, r, l);
				bsr(w2, x, l, shift);
				modmul(r, w1, w2, m, l);	/* r <- x . r */
			}
			b = (b << 1) & MASK;
		}
	}
}

/**********************************************************
 *                                                        *
 *       Modular exponentiation - small power             *
 *                                                        *
 **********************************************************/

void 
smallexp(r, x, p, m, l, shift, w1, w2)
	unsigned        r[];
	unsigned        x[];
	unsigned        p;
	unsigned        m[];
	int             l;
	int             shift;
	unsigned        w1[];
	unsigned        w2[];
{
	int             i;
	int             j;
	unsigned        b;
	unsigned        mask;

	mask = 1 << (BITS * 2 - 1);

	for (i = LSW; i <= l; i++)
		r[i] = 0;
	r[LSW] = 1 << shift;	/* r <- 1     */

	b = p;
	for (j = 0; j < BITS * 2; j++) {
		copy(w1, r, l);
		bsr(w2, r, l, shift);
		modmul(r, w1, w2, m, l);	/* r <- r . r */
		if ((b & mask) != 0) {
			copy(w1, r, l);
			bsr(w2, x, l, shift);
			modmul(r, w1, w2, m, l);	/* r <- x . r */
		}
		mask = mask >> 1;
	}
}

void 
fastexp(r, x, p, m, l, shift, w1, w2, table)
	unsigned        r[];	/* The result */
	unsigned        x[];
	unsigned        p[];	/* Power to use */
	unsigned        m[];	/* Modulus      */
	int             l;
	unsigned        w1[];	/* Workspace    */
	unsigned        w2[];
	unsigned      **table;

{
	int             i;
	int             j;
	unsigned        b;
	unsigned        cy;

	copy(w1, x, l);
	for (i = LSW; i <= l; i++) {
		copy(table[i], w1, l);
		cy = shl(w1, l);
		reduce(w1, cy, m, l);
	}
	clear(r, l);
	r[LSW] = 1 << shift;
	for (i = l; i >= LSW; i--) {
		b = p[i];
		for (j = 0; j < BITS; j++) {
			copy(w1, r, l);
			bsr(w2, r, l, shift);
			modmul(r, w1, w2, m, l);
			if (b & TOPB) {
				bsr(w1, r, l, shift);
				tablemul(r, w1, table, m, l);
			}
			b = (b << 1) & MASK;
		}
	}
}

/********************************************************
 *                                                      *
 *  Modular exponentiation, power is a Fermat number.   *
 *                                                      *
 *  3 is an obvious special case, for RSA. However, for *
 * reasons of security it is better to use 65537.       *
 *                                                      *
 ********************************************************/

void 
fermatexp(r, x, n, m, l, shift, w1, w2)
	unsigned        r[];	/* The result.              */
	unsigned        x[];	/* Number to raise to power */
	int             n;	/* Which Fermat number      */
	unsigned        m[];	/* The modulus              */
	int             l;	/* MSW of modulus           */
	int             shift;
	unsigned        w1[];	/* Workspace                */
	unsigned        w2[];	/* More workspace.          */
{
	int             i;

	copy(r, x, l);
	for (i = 0; i <= n; i++) {
		copy(w1, r, l);
		bsr(w2, r, l, shift);
		modmul(r, w1, w2, m, l);
	}
	copy(w1, r, l);
	bsr(w2, x, l, shift);
	modmul(r, w1, w2, m, l);
}
