;; Copyright (C) 1993 Fujitsu Laboratories Ltd.
;; written by Hiroshi TANAKA, tanaka@flab.fujitsu.co.jp
;;
;; FJPEM Setup Utility
;;

(defvar pem-makekey-command "/usr/local/bin/makekey"
  "*Makekey command for FJPEM.")

(defvar pem-makeinfo-command "/usr/local/bin/makeinfo"
  "*Makeinfo command for FJPEM.")

(defvar pem-sendmail-program
  (if (boundp 'sendmail-program) sendmail-program "/usr/lib/sendmail")
  "*Sendmail program for sendmail cert-request mail to CA.")

;;;(setq pem-ca-address (user-login-name))	;; for TEST
(defvar pem-ca-address "cert-req-flab@ca.fujitsu.co.jp"
  "*CA e-mail address for FJPEM.")

(setq pem-key-file ".keytest")	;; for TEST
(defvar pem-key-file ".key"
  "*Private-Key file-name for FJPEM.")

(setq pem-info-file ".infotest")	;; for TEST
(defvar pem-info-file ".info"
  "*Private-Key Info file-name for FJPEM.")

(defvar pem-save-info-p t
  "*Save Info file or not.")

(setq pem-info-list
  '(("N" nil nil)
    ("flab"   "flab$B>e$N#I#D(B" "kuroda")
    ("mail"   "$B%a!<%k%"%I%l%9(B" "kuroda@flab.fujitsu.co.jp")
    ("cn"     "$B;aL>!J1Q8l!K(B" "Yasutsugu Kuroda")
    ("kcn"    "$B;aL>!JF|K\8l!K(B" "$B9uED(B $B9/;L(B")
    ("title"  "$B8*=q$-!J1Q8l!K(B" "researcher")
    ("ktitle" "$B8*=q$-!JF|K\8l!K(B" "$B8&5f0w(B")
    ("telephoneNumber" "$BEEOCHV9f(B" "+81 44-754-2687")
    ("ou"     "$B=jB0!J1Q8l!K(B" "IP Network Center")
    ("kou"    "$B=jB0!JF|K\8l!K(B" "$B>pJs%7%9%F%`%;%s%?(B")))

(defvar pem-info-table (make-vector 10 0))

(defvar pem-conf nil)


;;
;; interfaces
;;
(defun pem-setup ()
  (interactive)
  (setq pem-conf (current-window-configuration))
  (condition-case ()
      (let ((key (expand-file-name (concat "~/" pem-key-file))))
	(if (file-exists-p key)
	    (if (y-or-n-p
		 (format "$BHkL)80%U%!%$%k!J(B%s$B!K$r=q$-49$($^$9$+!)(B" pem-key-file))
		(progn (call-process "/usr/bin/chmod" nil t nil "600" key)
		       (pem-makekey)))
	  (pem-makekey))
	(pem-makeinfo key)
	(pem-sendinfo)
	(pem-finish "PEM$B$N@_Dj$r=*$j$^$7$?!#(B"))
    (quit
     (ding)
     (pem-finish "PEM$B$N@_Dj$rCfCG$7$^$7$?!#(B"))
    (error
     (ding)
     (pem-finish "PEM$B$N@_Dj$O0[>o=*N;$7$^$7$?!#(B"))))

(defun pem-init ()
  (interactive)
  (let ((buffer (current-buffer)))
    (if (not (pem-cert-buffer-p buffer))
	(error "This Buffer is NOT a certification issue.")
      (pem-initcert buffer))))

;;
;; functions
;;
(defun pem-makekey ()
  (save-excursion
    (let ((rand (read-string "Input random string : "))
	  (buffer (get-buffer-create " *PEM Working Buffer* ")))
      (set-buffer buffer)
      (erase-buffer)
      (insert rand)
      (let ((default-directory (expand-file-name "~")))
	(call-process pem-makekey-command nil buffer nil pem-key-file)))))

(defun pem-makeinfo (key)
  (let ((buffer (get-buffer-create " *PEM Working Buffer* "))
	N flab-id mail)
    (pem-clear-info)
    (pem-restore-info)
    (save-excursion
      (set-buffer buffer)
      (erase-buffer)
      (call-process pem-makeinfo-command nil buffer nil key)
      (goto-char (point-min))
      (save-excursion
	(and (re-search-forward "^N= \\([^\n]+\\)" nil t)
	     (setq N (buffer-substring (match-beginning 1) (match-end 1)))))
      (save-excursion
	(and (re-search-forward "^flab= \\([^\n]+\\)" nil t)
	     (setq flab-id (buffer-substring (match-beginning 1) (match-end 1)))))
      (save-excursion
	(and (re-search-forward "^mail= \\([^\n]+\\)" nil t)
	     (setq mail (buffer-substring (match-beginning 1) (match-end 1))))))
    (pem-set-info "N" N)
    (pem-set-info "flab" flab-id)
    (pem-set-info "mail" mail)
    (pem-ask-info)
    (pem-save-info)))

(defun pem-sendinfo ()
  (let ((ilst pem-info-list)
	(buffer (get-buffer-create " *PEM Working Buffer* "))
	elm)
    (save-excursion
      (set-buffer buffer)
      (erase-buffer)
      (insert "From: " (pem-get-info "mail") "\n")
      (insert "To: " pem-ca-address "\n")
      (insert "Subject: FJPEM Certification Request --auto-generated mail by pem-setup.el--\n")
      (insert "\n")
      (while (setq elm (car ilst))
	(insert (car elm)
		"= "
		(pem-get-info (car elm))
		"\n")
	(setq ilst (cdr ilst)))
      (call-process-region (point-min) (point-max)
			   pem-sendmail-program nil t nil
			   pem-ca-address))
    ))

(defun pem-ask-info ()
  (let ((ilst pem-info-list)
	elm r)
    (while (setq elm (car ilst))
      (if (car (cdr elm))
	  (progn
	    (pem-pop-message (format "$B!ZNc![(B\n%s\n" (car (cdr (cdr elm)))))
	    (setq r (read-string (format "%s $B$rF~NO$7$F2<$5$$!'(B"
					 (car (cdr elm)))
				 (pem-get-info (car elm))))
	    (if (y-or-n-p (format "%s $B$G$$$$$G$9$+!)(B" r))
		(progn
		  (setq ilst (cdr ilst))
		  (pem-set-info (car elm) r))))
	(setq ilst (cdr ilst))))))

(defun pem-finish (msg)
  (pem-pop-message msg)
  (condition-case ()
      (read-string "$B!N(BReturn$B!O%-!<$rF~NO$7$F2<$5$$!#(B")
    (error nil)
    (quit nil))
  (set-window-configuration pem-conf))

(defun pem-pop-message (msg)
  (let ((buffer (get-buffer-create " *PEM Infomation* "))
	(window-min-height 2)
	lines)
    (delete-other-windows)
    (set-buffer buffer)
    (erase-buffer)
    (insert msg)
    (goto-char (point-min))
    (pop-to-buffer buffer)
    (setq lines (min (window-height)
		     (+ 2 (count-lines (point-min) (point-max)))))
    (enlarge-window (- lines (window-height)))
    (other-window 1)))

(defun pem-save-info ()
  (if (not pem-save-info-p)
      nil
    (let ((info (expand-file-name (concat "~/" pem-info-file)))
	  (ilst pem-info-list)
	  (buffer (get-buffer-create " *PEM Working Buffer* "))
	  elm)
      (save-excursion
	(set-buffer buffer)
	(erase-buffer)
	(insert "\n")
	(while (setq elm (car ilst))
	  (insert (car elm)
		  "= "
		  (pem-get-info (car elm))
		  "\n")
	  (setq ilst (cdr ilst)))
	(progn (and (file-exists-p info)
		    (call-process "/usr/bin/chmod" nil t nil "600" info))
	       (append-to-file (point-min) (point-max) info)
	       (call-process "/usr/bin/chmod" nil t nil "400" info))))))

(defun pem-restore-info ()
  (let ((info (expand-file-name (concat "~/" pem-info-file)))
	(ilst pem-info-list)
	(buffer (get-buffer-create " *PEM Working Buffer* "))
	elm)
    (if (not (file-exists-p info))
	nil
      (save-excursion
	(set-buffer buffer)
	(erase-buffer)
	(insert-file-contents info)
	(goto-char (point-max))
	(if (not (re-search-backward "^N=" nil t))
	    nil
	  (delete-region (point-min) (match-beginning 0))
	  (goto-char (point-min))
	  (while (setq elm (car ilst))
	    (save-excursion
	      (and (re-search-forward
		    (format "^%s= \\([^\n]+\\)" (car elm)) nil t)
		   (pem-set-info
		    (car elm)
		    (buffer-substring (match-beginning 1) (match-end 1)))))
	    (setq ilst (cdr ilst))))))))

(defun pem-clear-info ()
  (setq pem-info-table (make-vector 10 0)))

(defun pem-set-info (key val)
  (let ((sym (intern key pem-info-table)))
    (put sym 'val val)
    (put sym 'key key)))

(defun pem-get-info (key)
  (let ((sym (intern-soft key pem-info-table)))
    (and sym (get sym 'val))))

