/* dpmlib.c     COPYRIGHT Fujitsu Laboratories Ltd. 1993 */
/* $Id: dpmlib.c,v 1.1 1995/03/08 06:04:33 kuroda Exp $ */
/*
***	dpem.c - decrypt PEM message 	Nov 20,92 /kikn
***	dpmlib.c 		revised Aug. 30, 93 /kikn
 */

#include	<stdio.h>
#include	"pem.h"
#include	"cert.h"
#include	"rsa.h"

depem(mycert, mykey, myparms)
BER			*mycert;
struct RSASecretKey	*mykey;
struct RSAParameters	*myparms;
{
	static char		tag[BUFSIZ], args[BUFSIZ]; 
	static unsigned char	dek[BUFSIZ], iv[BUFSIZ], iv2[BUFSIZ];
	static unsigned char	user[BUFSIZ];
	static unsigned char	mode[BUFSIZ], domain[BUFSIZ];
	static unsigned	char	mic[BUFSIZ];
	static unsigned	char	md[BUFSIZ], buf[BUFSIZ];

	BER	*sender, *issuer, *ca, *mdinfo, *myissuer;
	
	struct	RSAPublicKey	spub, ipub;
	struct	RSAParameters	sparms, iparms;
	int	recflag = 0, id = 0;
	int	i, ver, j;
	int	miclen = 0, havedek = 0;
	char	*errstr, *s;
	int	micerr = 0;

	seekbound();
	while(gettag(tag, args) != NULL){
		if(strcmp(tag, "Proc-Type") == 0){
			getver(args, mode);
		} else if(strcmp(tag, "Content-Domain") == 0){
			getdomain(args, domain);
		} else if(strcmp(tag, "DEK-Info") == 0){
			getiv(args, iv);
		} else if(strcmp(tag, "Originator-Certificate") == 0){
			sender = getber();
			ca = C_IN(sender);
			id = SerialNumber(sender);
		} else if(strcmp(tag, "Issuer-Certificate") == 0){
			issuer = getber();
		} else if(strcmp(tag, "MIC-Info") == 0){
			miclen = getmic(args, mic);
		} else if(strcmp(tag, "Recipient-ID-Asymmetric") == 0){
			ca = getber();
			fscanf(fin, " %x\n", &id);
		} else if(strcmp(tag, "Key-Info") == 0){
			if(verbose){ fprintf(stderr, "ID = %x\nCID = %x\n", id, SerialNumber(mycert)); }
			if(!havedek && cmpber(ca, C_IN(mycert)) == 0
			&& id == (int)SerialNumber(mycert)){
				getkey(args, dek, mykey, myparms);
				havedek = 1;
			} else {
				freadfield(fin, buf);
			}
		} else {
			fprintf(stderr, "unaccepteable header %s\n", tag);
			exit(1);
		}
	}

	if(strcmp(mode, "MIC-CLEAR") == 0){
		md5line(md, fin);
		getpubkey(C_PK(sender), &spub, &sparms);
		dek[0] = 0;
		mdinfo = micinfo(mic, miclen, dek, iv, &spub, &sparms);
	}else if(strcmp(mode, "ENCRYPTED") == 0){
		if(!integrity){
			decmail(dek, iv);	
			return(0);
		}else{
			bcopy(iv, iv2, DEKSIZE);
			decmailmd5(dek, iv, md);	
			getpubkey(C_PK(sender), &spub, &sparms);
			mdinfo = micinfo(mic, miclen, dek, iv2, &spub, &sparms); 
		}
	}
	if(verbose){
		micreport(md, mdinfo);
	}
	if(!integrity){
		return(0);
	}

	fputs("----- FJPEM INTEGRITY REPORT -----\n", fout);
	fputs("Message Integrity: ", fout);
	if(mdinfo == NULL){
		fputs("INVALID\n\tBad Encryption\n", fout);
	}else if(bcmp(md, CDRS(mdinfo)[1]->val, MICSIZE)){
		fputs("INVALID\n", fout);
		micreport(md, mdinfo);
	}else{
		fputs("Valid\n", fout);
	}
	getpubkey(C_PK(issuer), &ipub, &iparms);
	fputs("Originator Certificate: ", fout);
	errstr = vfysig(sender, &ipub, &iparms);
	if(errstr){
		micerr = 3;
		fputs("INVALID\n", fout);
		fprintf(fout, "\tWARNING (%s)\n", errstr);
		prnname(C_SN(sender));
		fprintf(fout, "\tSerial No = %X\n", SerialNumber(sender));
		fprintf(fout, "\tGood %s through %s\n", 
			ValidityFrom(sender), ValidityTo(sender)); 
	}else{
		fputs("Valid\n", fout);
		prnname(C_SN(sender));
	}
	if(mymail){
		if(strcasecmp(mymail, (s = getmail(C_SN(sender))))){
			fprintf(fout, "\tWARNING: FROM(%s)\n", mymail);
			fprintf(fout, "\tWARNING: CERT(%s)\n", s);
			micerr = 2;
		}
	}
	if(issuercert){
		fputs("Issuer Certificate: ", fout);
		myissuer = readcert(issuercert);
		if(cmpber(myissuer, issuer)){
			fputs("UNKNOWN CA\n", fout);
			fprintf(fout, "\tSerial No = %X\n", SerialNumber(issuer));
			prnname(C_SN(issuer));
			micerr = 1;
		}else{
			fputs("Valid\n", fout);
		}
	}
	fputs("----- FJPEM REPORT END -----\n", fout);
	return(micerr);
}
		
micreport(md, mdinfo)
unsigned char	md[], mdinfo[];
{
	int i;

	fprintf(fout, "\tMICCALC: ");
	for(i = 0; i < MICSIZE; ++i){
		fprintf(fout, "%x", md[i]);
	}
	if(mdinfo[0] == 0x30){
		dispber(mdinfo, "\n\tMICINFO:\t");
		fputs("\n", fout);
	} else {
		fprintf(fout, "Bad encryption\n\t" );
		for(i = 0; i < MICSIZE; ++i){
			fprintf(fout, "%x", mdinfo[i]);
		}
	}
}

gettag(tag, args)
char	tag[], args[];
{
	static char	buf[BUFSIZ];

	if(fgets(buf, BUFSIZ, fin) == NULL){
		fprintf(stderr, "missing headers\n");
		exit(1);
	}
	if(strcmp(buf,"\n") == 0){
		return(NULL);
	}
	sscanf(buf, "%[^:]: %[^\n]\n", tag, args);
	if(verbose){
		fprintf(stderr, "%s:\n", tag);
	}
	return(1);
}

BER*
getber()
{
	unsigned char	buf[BUFSIZ];

	freadfield(fin, buf);
	return(deber(buf));
}


getdomain(args, domain)
char	args[], domain[];
{
	sscanf(args, "%s\n", domain);
	if(strcmp(domain, "RFC822")){
		fprintf(stderr, "Unknown domain %s\n", domain);
		exit(1);
	}
	if(verbose)
	fprintf(stderr, "\tDomain = %s\n", domain);
}

getver(args, mode)
char	args[], mode[];
{
	int	ver = 0;

	sscanf(args, "%d,%s\n", &ver, mode);
	if(ver != 4){
		fprintf(stderr, "unacceptable version %d\n", ver);
	}	
	if(verbose)
	fprintf(stderr, "\tVersion = %d, Mode = %s\n", ver, mode);
}

getiv(args, iv)
char 		args[];
unsigned char	iv[];
{
	char	mode[BUFSIZ], dek[BUFSIZ];
	int	c, i;

	sscanf(args, " %[^,],%s\n", mode, dek);
	for(i = 0; i < DEKSIZE; ++i){
		sscanf(&dek[i * 2], "%02X", &c);
		iv[i] = c;
	}

	if(strcmp(mode, "DES-CBC")){
		fprintf(stderr, "unacceptable mode %s\n", mode);
	}
	if(verbose){
		fprintf(stderr, "\tIV = ");
		for(i = 0; i < DEKSIZE; ++i){
			fprintf(stderr, "%02X", iv[i]);
		}
	}
}


getmic(args, mic)
char		args[];
unsigned char	mic[];
{
	static char	md[BUFSIZ], alg[BUFSIZ];

	sscanf(args, "%[^,],%[^,],\n", md, alg);
	if(verbose)
	fprintf(stderr, "\tMD Algorithm = %s, Signature Algorithm = %s, ",
		md, alg);

	return(freadfield(fin, mic));	
}


seekbound()
{
	static char	buf[BUFSIZ];
	int	s = strlen(BEGIN_EB);

	while(fgets(buf, BUFSIZ, fin) != NULL){
		if(strncmp(buf, BEGIN_EB, s) == 0){
			return(0);
		}
		fputs(buf, fout);
	}
	exit(0);
	/* exit(1) "no bound" */
}
