/* miclib.c     COPYRIGHT Fujitsu Laboratories Ltd. 1993 */
/* $Id: miclib.c,v 1.1 1995/03/08 06:04:38 kuroda Exp $ */
/*
***	miclib.c - message integrity check on sender's IK Nov 20,92 /kikn
***			(modified	Aug. 3,92)
 */

#include	<stdio.h>
#include 	"pem.h"
#include	"rsa.h"
#include 	"MD/global.h"
#include 	"MD/md5.h"
#include 	"MD/alg_md2.h"


/* template for BER encoding */
unsigned char	BERmd5[64] = {
	0x30, 0x20, 0x30, 0x0c, 0x06, 0x08, 0x2a, 0x86, 0x48,
	0x86, 0xf7, 0x0d, 0x02, 0x05, 0x05, 0x00, 0x04, 0x10,
	0x12, 0x4f, 0x9b, 0x49, 0xce, 0xa6, 0x2e, 0x26, 0x67,
	0x04, 0xc0, 0x5b, 0xab, 0x10, 0xbc, 0xea
};

putmic(md, dek, iv, key, parms)
unsigned char	md[], dek[], iv[];
struct RSASecretKey	*key;
struct RSAParameters	*parms;
{
	static unsigned char	mic[BUFSIZ], mic2[BUFSIZ], mic3[BUFSIZ];
	static unsigned char	mdber[BUFSIZ];
	static unsigned		w[BUFSIZ], *w2;
	int	c, i;

	c = putmd5(md, mdber);
	if(verbose){
		for(i = 0; i < c; ++i){ 
			fprintf(stderr, "%.02x", mdber[i]);
		}
		fprintf(stderr, "\n");
	}
	pad(1, mdber, c, mic, parms->len * 2);

	char2word(mic, w, parms->len * 2);
	rsa_decrypt(w, &w2, key, parms);
	word2char(w2, mic2, parms->len);
	if(verbose){
		fhexdump(stderr, w, 31);
		fhexdump(stderr, w2, 31);
		for(i = 0; i < 64; ++i){ 
			fprintf(stderr, "%.02x", mic2[i]);
		}
		fprintf(stderr, "\n");
	}

	if(dek[0] == 0){
		writebyte(mic2, parms->len * 2);
		writebyte(mic2, 0);
	}else{
		c = encdes(mic2, mic3, parms->len * 2, dek, iv);
		writebyte(mic3, c);
		writebyte(mic3, 0);
	}
}


BER*
micinfo(mic, miclen, dek, iv, pub, parms)
unsigned char	mic[];
int		miclen;
unsigned char	dek[], iv[];
struct	RSAPublicKey	*pub;
struct	RSAParameters	*parms;
{
	int	c;
	static unsigned char	m2[BUFSIZ], m3[BUFSIZ], md[BUFSIZ];
	static unsigned 	w[BUFSIZ], *w2;

	if(dek[0]){
		c = dedes(mic, m2, miclen, dek, iv);	
	}else{
		bcopy(mic, m2, miclen);
		c = miclen;
	}

	char2word(m2, w, c);
	rsa_encrypt(w, &w2, pub, parms);
	word2char(w2, m3, parms->len);

	padout(m3, md, MDSIZE + 10);
	/* for(c = 0; c < 30; ++c) fprintf(stderr, "%02X", md[c]); */
	if(md[0] != 0x30 || md[2] != 0x30){
		return(NULL);
	}
	return(deber(md));
}


putmd5(mic, buf)
unsigned char	mic[], buf[];
{
	BER	*ber;

	ber = deber(BERmd5);
	bcopy(mic, ((BER **)ber->val)[1]->val, MICSIZE);
	enber(ber, buf);
	/*	dispber(ber, "\nMIC:");
		printf("\n\n");	*/
	return(ber->byte);
}

md5line(mic, fp)	/* called in a verification of MIC-CLEAR only */
unsigned char 	mic[];
FILE		*fp;
{
	unsigned char	buf[BUFSIZ], net[BUFSIZ];
	MD5_CTX context;
	int 	len, ofs;

	bzero(mic, MICSIZE);
	MD5Init(&context);
	while((fgets(buf, BUFSIZ, fp) != NULL) && (strncmp(buf, "--", 2) )){
		ofs = (buf[0] == '-')? 2 : 0;
		fputs(&buf[ofs], fout);
		len = crlfn(&buf[ofs], net, strlen(&buf[ofs]));
		MD5Update(&context, net, len);
	}
	MD5Final(mic, &context);
}

md5(mic, fp)	/* in signing a text in MIC-CLEAR */
unsigned char 	mic[];
FILE		*fp;
{
	unsigned char	buf[BUFSIZ], net[BUFSIZ];
	MD5_CTX context;
	int 	len, len2;
	static unsigned char	crlf[2] = {0x0d, 0x0a};

	bzero(mic, MICSIZE);
	MD5Init(&context);
	while(fgets(buf, BUFSIZ, fp) != NULL){
		fputs(buf, fout);
		len = crlfn(buf, net, strlen(buf));
		MD5Update(&context, net, len);
	}
	if(net[len -1] != 0x0a){
		fputc(0x0a, fout);
		MD5Update(&context, crlf, 2);
	}
	MD5Final(mic, &context);
}

md2(mic, in, len)
unsigned char 	mic[], in[];
int	len;
{
	unsigned char	buf[BUFSIZ];
	struct	MD2Hash *hash;
	MD5_CTX context;
	int	blocks, extra, i;

	bzero(mic, BUFSIZ);
	md2_init_hash(&hash, NULL, NULL);

	blocks = len /16;
	extra = len % 16;
	for(i = 0; i < blocks; ++i){
		md2_hash_block(hash, &in[i * 16], NULL, NULL);
	}
	bzero(buf, BUFSIZ);
	bcopy(&in[blocks * 16], buf, extra);
	md2_hash_short_block(hash, buf, extra, NULL, NULL);
	md2_terminate_hash(hash, NULL, NULL);
	for(i = 0; i < 16; ++i){
		mic[i] = hash->D[i] & 0xff;
	}
}


